import datetime
import time

from django import forms
from django.utils.translation import ugettext_lazy as _

import stripe
from ipware.ip import get_ip, get_real_ip

from .actions import accounts
from .conf import settings
from .models import Plan


class PaymentMethodForm(forms.Form):

    expMonth = forms.IntegerField(min_value=1, max_value=12)
    expYear = forms.IntegerField(min_value=2015, max_value=9999)


class PlanForm(forms.Form):
    plan = forms.ModelChoiceField(queryset=Plan.objects.all())


"""
The Connect forms here are designed to get users through the multi-stage
verification process Stripe uses for custom accounts, as detailed here:

https://stripe.com/docs/connect/testing-verification

You can view the required fields on a per-country basis using the API:

https://stripe.com/docs/api#country_spec_object

The following forms are sufficient for the US and Canada.
"""

# Note: undocumented, determined through experimentation
STRIPE_MINIMUM_DOB = datetime.date(1900, 1, 1)


ACCEPTED_DOCUMENT_CONTENT_TYPES = (
    "image/jpg", "image/jpeg", "image/png"
)

COUNTRY_CHOICES = [
    ("CA", _("Canada")),
    ("US", _("United States"))
]

STATE_CHOICES_BY_COUNTRY = {
    "CA": [
        ("AB", _("Alberta")),
        ("BC", _("British Columbia")),
        ("MB", _("Manitoba")),
        ("NB", _("New Brunswick")),
        ("NL", _("Newfoundland and Labrador")),
        ("NT", _("Northwest Territories")),
        ("NS", _("Nova Scotia")),
        ("NU", _("Nunavut")),
        ("ON", _("Ontario")),
        ("PE", _("Prince Edward Island")),
        ("QC", _("Quebec")),
        ("SK", _("Saskatchewan")),
        ("YT", _("Yukon"))
    ],
    "US": [
        ("AL", _("Alabama")),
        ("AK", _("Alaska")),
        ("AZ", _("Arizona")),
        ("AR", _("Arkansas")),
        ("CA", _("California")),
        ("CO", _("Colorado")),
        ("CT", _("Connecticut")),
        ("DE", _("Delaware")),
        ("DC", _("District of Columbia")),
        ("FL", _("Florida")),
        ("GA", _("Georgia")),
        ("HI", _("Hawaii")),
        ("ID", _("Idaho")),
        ("IL", _("Illinois")),
        ("IN", _("Indiana")),
        ("IA", _("Iowa")),
        ("KS", _("Kansas")),
        ("KY", _("Kentucky")),
        ("LA", _("Louisiana")),
        ("ME", _("Maine")),
        ("MD", _("Maryland")),
        ("MA", _("Massachusetts")),
        ("MI", _("Michigan")),
        ("MN", _("Minnesota")),
        ("MS", _("Mississippi")),
        ("MO", _("Missouri")),
        ("MT", _("Montana")),
        ("NE", _("Nebraska")),
        ("NV", _("Nevada")),
        ("NH", _("New Hampshire")),
        ("NJ", _("New Jersey")),
        ("NM", _("New Mexico")),
        ("NY", _("New York")),
        ("NC", _("North Carolina")),
        ("ND", _("North Dakota")),
        ("OH", _("Ohio")),
        ("OK", _("Oklahoma")),
        ("OR", _("Oregon")),
        ("PA", _("Pennsylvania")),
        ("RI", _("Rhode Island")),
        ("SC", _("South Carolina")),
        ("SD", _("South Dakota")),
        ("TN", _("Tennessee")),
        ("TX", _("Texas")),
        ("UT", _("Utah")),
        ("VT", _("Vermont")),
        ("VA", _("Virginia")),
        ("WA", _("Washington")),
        ("WV", _("West Virginia")),
        ("WI", _("Wisconsin")),
        ("WY", _("Wyoming"))
    ]
}

CURRENCY_CHOICES_BY_COUNTRY = {
    "CA": [
        ("CAD", _("CAD: Canadian Dollars")),
        ("USD", _("USD: US Dollars")),
    ],
    "US": [
        ("USD", _("USD: US Dollars")),
    ]
}

FIELDS_BY_COUNTRY = {
    "CA": {
        "legal_entity.personal_id_number": (
            "personal_id_number",
            forms.CharField(
                label=_("SIN")
            ),
        ),
        "legal_entity.verification.document": (
            "document",
            forms.FileField(
                label=_("Scan of government-issued ID")
            ),
        )
    },
    "US": {
        "legal_entity.personal_id_number": (
            "personal_id_number",
            forms.CharField(
                label=_("SSN")
            )
        ),
        "legal_entity.verification.document": (
            "document",
            forms.FileField(
                label=_("Scan of government-issued ID")
            ),
        )
    }
}

# lookup local form fields for Stripe field errors
# we use `contains` so the stripe side (left) need
# not be super specific

STRIPE_FIELDS_TO_LOCAL_FIELDS = {
    "dob": "dob",
    "first_name": "first_name",
    "second_name": "second_name",
    "routing_number": "routing_number",
    "currency": "currency",
    "account_number": "account_number",
    "personal_id_number": "personal_id_number",
    "file": "document"
}


class DynamicManagedAccountForm(forms.Form):
    """Set up fields according to fields needed and relevant country."""

    def __init__(self, *args, **kwargs):
        self.country = kwargs.pop("country")
        self.fields_needed = kwargs.pop("fields_needed", [])
        super(DynamicManagedAccountForm, self).__init__(*args, **kwargs)
        # build our form using the country specific fields and falling
        # back to our default set
        for f in self.fields_needed:
            if f in FIELDS_BY_COUNTRY.get(self.country, {}):  # pragma: no branch
                field_name, field = FIELDS_BY_COUNTRY[self.country][f]
                self.fields[field_name] = field

    # clean methods only kick in if the form has the relevant field

    def clean_document(self):
        document = self.cleaned_data.get("document")
        if document._size > settings.PINAX_STRIPE_DOCUMENT_MAX_SIZE_KB:
            raise forms.ValidationError(
                _("Document image is too large (> %(maxsize)s MB)") % {
                    "maxsize": round(
                        float(
                            settings.PINAX_STRIPE_DOCUMENT_MAX_SIZE_KB
                        ) / float(
                            1024 * 1024
                        ), 1
                    )
                }
            )
        if document.content_type not in ACCEPTED_DOCUMENT_CONTENT_TYPES:
            raise forms.ValidationError(
                _(
                    "The type of image you supplied is not supported. "
                    "Please upload a JPG or PNG file."
                )
            )
        return document

    def clean_dob(self):
        data = self.cleaned_data["dob"]
        if data < STRIPE_MINIMUM_DOB:
            raise forms.ValidationError(
                "This must be greater than {}.".format(
                    STRIPE_MINIMUM_DOB
                )
            )
        return data

    def stripe_field_to_local_field(self, stripe_field):
        for r, l in STRIPE_FIELDS_TO_LOCAL_FIELDS.items():
            if r in stripe_field:
                return l

    def stripe_error_to_form_error(self, error):
        """
        Translate a Stripe error into meaningful form feedback.

        error.json_body = {
            u'error': {
                u'message':
                u"This value must be greater than 1900.",
                u'type': u'invalid_request_error',
                u'param': u'legal_entity[dob][year]'
            }
        }
        """
        message = error.json_body["error"]["message"]
        stripe_field = error.json_body["error"]["param"]
        local_field = self.stripe_field_to_local_field(stripe_field)
        if local_field:
            self.add_error(local_field, message)
        else:
            self.add_error(None, message)


def extract_ipaddress(request):
    """Extract IP address from request."""
    ipaddress = get_real_ip(request)
    if not ipaddress and settings.DEBUG:  # pragma: no cover
        ipaddress = get_ip(request)
    return ipaddress


class InitialCustomAccountForm(DynamicManagedAccountForm):
    """
    Collect `minimum` fields for CA and US CountrySpecs.

    Note: for US, `legal_entity.ssn_last_4` appears in the `minimum`
    set but in fact is not required for the account to be functional.
    Similarly for CA, `legal_entity.personal_id_number` is listed as
    `minimum` but in practice is not required to be able to charge
    and transfer.
    """

    first_name = forms.CharField(max_length=100)
    last_name = forms.CharField(max_length=100)
    dob = forms.DateField()

    address_line1 = forms.CharField(max_length=300)
    address_city = forms.CharField(max_length=100)
    address_state = forms.CharField(max_length=100)
    address_country = forms.ChoiceField(choices=COUNTRY_CHOICES)
    address_postal_code = forms.CharField(max_length=100)

    # for external_account
    routing_number = forms.CharField(max_length=100)
    account_number = forms.CharField(max_length=100)

    tos_accepted = forms.BooleanField()

    def __init__(self, *args, **kwargs):
        """Instantiate no fields based on `fields_needed` initially."""
        self.request = kwargs.pop("request")
        super(InitialCustomAccountForm, self).__init__(
            *args, **kwargs
        )
        self.fields["address_state"] = forms.ChoiceField(
            choices=STATE_CHOICES_BY_COUNTRY[self.country]
        )
        self.fields["currency"] = forms.ChoiceField(
            choices=CURRENCY_CHOICES_BY_COUNTRY[self.country]
        )

    def get_ipaddress(self):
        return extract_ipaddress(self.request)

    def get_user_agent(self):
        return self.request.META.get("HTTP_USER_AGENT")

    def save(self):
        """
        Create a custom account, handling Stripe errors.

        Note: the below will create a custom, manually paid out
        account. This is here mostly as an example, you will likely
        need to override this method and do your own application
        specific special sauce.
        """
        data = self.cleaned_data
        try:
            return accounts.create(
                self.request.user,
                country=data["address_country"],
                type="custom",
                legal_entity={
                    "dob": {
                        "day": data["dob"].day,
                        "month": data["dob"].month,
                        "year": data["dob"].year
                    },
                    "first_name": data["first_name"],
                    "last_name": data["last_name"],
                    "type": "individual",
                    "address": {
                        "line1": data["address_line1"],
                        "city": data["address_city"],
                        "postal_code": data["address_postal_code"],
                        "state": data["address_state"],
                        "country": data["address_country"]
                    }
                },
                tos_acceptance={
                    "date": int(time.time()),
                    "ip": self.get_ipaddress(),
                    "user_agent": self.get_user_agent()
                },
                transfer_schedule={
                    "interval": "manual"
                },
                external_account={
                    "object": "bank_account",
                    "account_holder_name": u" ".join(
                        [
                            data["first_name"],
                            data["last_name"]
                        ]
                    ),
                    "country": data["address_country"],
                    "currency": data["currency"],
                    "account_holder_type": "individual",
                    "default_for_currency": True,
                    "account_number": data["account_number"],
                    "routing_number": data["routing_number"]
                },
                # useful reference to our local user instance
                metadata={
                    "user_id": self.request.user.id
                }
            )

        except stripe.error.InvalidRequestError as se:
            self.stripe_error_to_form_error(se)
            raise


class AdditionalCustomAccountForm(DynamicManagedAccountForm):
    """
    Collect `additional` fields for CA and US CountrySpecs.

    Note: for US, `legal_entity.ssn_last_4` appears in the `minimum`
    set but in fact is not required for the account to be functional.
    Similarly for CA, `legal_entity.personal_id_number` is listed as
    `minimum` but in practice is not required to be able to charge
    and transfer.

    It's possible when further verification is needed that the user
    made a mistake with their name or dob, so we include these
    fields so the user can make any adjustments.
    """

    first_name = forms.CharField(max_length=100)
    last_name = forms.CharField(max_length=100)
    dob = forms.DateField()

    def __init__(self, *args, **kwargs):
        self.account = kwargs.pop("account")
        kwargs.update(
            {
                "country": self.account.country,
                "fields_needed": self.account.verification_fields_needed,
            }
        )
        super(AdditionalCustomAccountForm, self).__init__(*args, **kwargs)
        # prepopulate with the existing account details
        self.fields["first_name"].initial = self.account.legal_entity_first_name
        self.fields["last_name"].initial = self.account.legal_entity_last_name
        self.fields["dob"].initial = self.account.legal_entity_dob

    def save(self):
        data = self.cleaned_data
        try:
            return accounts.update(
                self.account,
                {
                    "dob": {
                        "day": data["dob"].day,
                        "month": data["dob"].month,
                        "year": data["dob"].year
                    },
                    "first_name": data["first_name"],
                    "last_name": data["last_name"],
                    "personal_id_number": data.get("personal_id_number"),
                    "document": data.get("document")
                }
            )
        except stripe.error.InvalidRequestError as se:
            self.stripe_error_to_form_error(se)
            raise



import stripe

from .. import models, utils


def during(year, month):
    """
    Return a queryset of pinax.stripe.models.Transfer objects for the provided
    year and month.

    Args:
        year: 4-digit year
        month: month as a integer, 1=January through 12=December
    """
    return models.Transfer.objects.filter(
        date__year=year,
        date__month=month
    )


def sync_transfer(transfer, event=None):
    """
    Synchronize a transfer from the Stripe API

    Args:
        transfer: data from Stripe API representing transfer
        event: the event associated with the transfer
    """
    defaults = {
        "amount": utils.convert_amount_for_db(
            transfer["amount"], transfer["currency"]
        ),
        "amount_reversed": utils.convert_amount_for_db(
            transfer["amount_reversed"], transfer["currency"]
        ) if transfer.get("amount_reversed") else None,
        "application_fee": utils.convert_amount_for_db(
            transfer["application_fee"], transfer["currency"]
        ) if transfer.get("application_fee") else None,
        "created": utils.convert_tstamp(transfer["created"]) if transfer.get("created") else None,
        "currency": transfer["currency"],
        "date": utils.convert_tstamp(transfer.get("date")),
        "description": transfer.get("description"),
        "destination": transfer.get("destination"),
        "destination_payment": transfer.get("destination_payment"),
        "event": event,
        "failure_code": transfer.get("failure_code"),
        "failure_message": transfer.get("failure_message"),
        "livemode": transfer.get("livemode"),
        "metadata": dict(transfer.get("metadata", {})),
        "method": transfer.get("method"),
        "reversed": transfer.get("reversed"),
        "source_transaction": transfer.get("source_transaction"),
        "source_type": transfer.get("source_type"),
        "statement_descriptor": transfer.get("statement_descriptor"),
        "status": transfer.get("status"),
        "transfer_group": transfer.get("transfer_group"),
        "type": transfer.get("type")
    }
    obj, created = models.Transfer.objects.update_or_create(
        stripe_id=transfer["id"],
        defaults=defaults
    )
    if not created:
        obj.status = transfer["status"]
        obj.save()
    return obj


def update_status(transfer):
    """
    Update the status of a pinax.stripe.models.Transfer object from Stripe API

    Args:
        transfer: a pinax.stripe.models.Transfer object to update
    """
    transfer.status = stripe.Transfer.retrieve(transfer.stripe_id).status
    transfer.save()


def create(amount, currency, destination, description, transfer_group=None,
           stripe_account=None, **kwargs):
    """
    Create a transfer.

    Args:
        amount: quantity of money to be sent
        currency: currency for the transfer
        destination: stripe_id of either a connected Stripe Account or Bank Account
        description: an arbitrary string displayed in the webui alongside the transfer
        transfer_group: a string that identifies this transfer as part of a group
        stripe_account: the stripe_id of a Connect account if creating a transfer on
            their behalf
    """
    kwargs.update(dict(
        amount=utils.convert_amount_for_api(amount, currency),
        currency=currency,
        destination=destination,
        description=description
    ))
    if transfer_group:
        kwargs["transfer_group"] = transfer_group
    if stripe_account:
        kwargs["stripe_account"] = stripe_account
    transfer = stripe.Transfer.create(**kwargs)
    return sync_transfer(transfer)



from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.contrib.auth import get_user_model
from django.db.models import Count
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

from .models import (
    Account,
    BankAccount,
    BitcoinReceiver,
    Card,
    Charge,
    Coupon,
    Customer,
    Event,
    EventProcessingException,
    Invoice,
    InvoiceItem,
    Plan,
    Subscription,
    Transfer,
    TransferChargeFee,
    UserAccount
)


def user_search_fields():
    User = get_user_model()
    fields = [
        "user__{0}".format(User.USERNAME_FIELD)
    ]
    if "email" in [f.name for f in User._meta.fields]:  # pragma: no branch
        fields += ["user__email"]
    return fields


def customer_search_fields():
    return [
        "customer__{0}".format(field)
        for field in user_search_fields()
    ]


class CustomerHasCardListFilter(admin.SimpleListFilter):
    title = "card presence"
    parameter_name = "has_card"

    def lookups(self, request, model_admin):
        return [
            ["yes", "Has Card"],
            ["no", "Does Not Have a Card"]
        ]

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(card__isnull=True)
        elif self.value() == "no":
            return queryset.filter(card__isnull=False)
        return queryset.all()


class InvoiceCustomerHasCardListFilter(admin.SimpleListFilter):
    title = "card presence"
    parameter_name = "has_card"

    def lookups(self, request, model_admin):
        return [
            ["yes", "Has Card"],
            ["no", "Does Not Have a Card"]
        ]

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(customer__card__isnull=True)
        elif self.value() == "no":
            return queryset.filter(customer__card__isnull=False)
        return queryset.all()


class CustomerSubscriptionStatusListFilter(admin.SimpleListFilter):
    title = "subscription status"
    parameter_name = "sub_status"

    def lookups(self, request, model_admin):
        statuses = [
            [x, x.replace("_", " ").title()]
            for x in Subscription.objects.all().values_list(
                "status",
                flat=True
            ).distinct()
        ]
        statuses.append(["none", "No Subscription"])
        return statuses

    def queryset(self, request, queryset):
        if self.value() == "none":
            # Get customers with 0 subscriptions
            return queryset.annotate(subs=Count("subscription")).filter(subs=0)
        elif self.value():
            # Get customer pks without a subscription with this status
            customers = Subscription.objects.filter(
                status=self.value()).values_list(
                "customer", flat=True).distinct()
            # Filter by those customers
            return queryset.filter(pk__in=customers)
        return queryset.all()


class AccountListFilter(admin.SimpleListFilter):
    title = "account"
    parameter_name = "stripe_account"

    def lookups(self, request, model_admin):
        return [("none", "Without Account")] + [(a.pk, str(a)) for a in Account.objects.all()]

    def queryset(self, request, queryset):
        if self.value() == "none":
            return queryset.filter(stripe_account__isnull=True)
        if self.value():
            return queryset.filter(stripe_account__pk=self.value())
        return queryset


class PrefetchingChangeList(ChangeList):
    """A custom changelist to prefetch related fields."""
    def get_queryset(self, request):
        qs = super(PrefetchingChangeList, self).get_queryset(request)

        if subscription_status in self.list_display:
            qs = qs.prefetch_related("subscription_set")
        if "customer" in self.list_display:
            qs = qs.prefetch_related("customer")
        if "user" in self.list_display:
            qs = qs.prefetch_related("user")
        return qs


class ModelAdmin(admin.ModelAdmin):
    def has_add_permission(self, request, obj=None):
        return False

    def change_view(self, request, object_id, form_url="", extra_context=None):
        """Adjust change_view title ("View" instead of "Change")."""
        opts = self.model._meta

        extra_context = extra_context or {}
        extra_context["title"] = _("View %s" % force_text(opts.verbose_name))
        return super(ModelAdmin, self).change_view(
            request, object_id, form_url, extra_context=extra_context,
        )

    def has_change_permission(self, request, obj=None):
        if request.method == "POST":
            return False
        return True

    def get_changelist(self, request, **kwargs):
        return PrefetchingChangeList


class ChargeAdmin(ModelAdmin):
    list_display = [
        "stripe_id",
        "customer",
        "total_amount",
        "description",
        "paid",
        "disputed",
        "refunded",
        "receipt_sent",
        "created_at",
    ]
    list_select_related = [
        "customer",
    ]
    search_fields = [
        "stripe_id",
        "customer__stripe_id",
        "invoice__stripe_id",
    ] + customer_search_fields()
    list_filter = [
        "paid",
        "disputed",
        "refunded",
        "created_at",
    ]
    raw_id_fields = [
        "customer",
        "invoice",
    ]
    readonly_fields = [
        "stripe_account_stripe_id",
    ]

    def get_queryset(self, request):
        qs = super(ChargeAdmin, self).get_queryset(request)
        return qs.prefetch_related("customer__user", "customer__users")


class EventProcessingExceptionAdmin(ModelAdmin):
    list_display = [
        "message",
        "event",
        "created_at"
    ]
    search_fields = [
        "message",
        "traceback",
        "data"
    ]
    raw_id_fields = [
        "event"
    ]


class EventAdmin(ModelAdmin):
    raw_id_fields = ["customer", "stripe_account"]
    list_display = [
        "stripe_id",
        "kind",
        "livemode",
        "valid",
        "processed",
        "created_at",
        "stripe_account",
    ]
    list_filter = [
        "kind",
        "created_at",
        "valid",
        "processed",
        AccountListFilter,
    ]
    search_fields = [
        "stripe_id",
        "customer__stripe_id",
        "validated_message",
        "=stripe_account__stripe_id",
    ] + customer_search_fields()


class SubscriptionInline(admin.TabularInline):
    model = Subscription
    extra = 0
    max_num = 0


class CardInline(admin.TabularInline):
    model = Card
    extra = 0
    max_num = 0


class BitcoinReceiverInline(admin.TabularInline):
    model = BitcoinReceiver
    extra = 0
    max_num = 0


def subscription_status(obj):
    return ", ".join([subscription.status for subscription in obj.subscription_set.all()])
subscription_status.short_description = "Subscription Status"  # noqa


class CustomerAdmin(ModelAdmin):
    raw_id_fields = ["user", "stripe_account"]
    list_display = [
        "stripe_id",
        "user",
        "account_balance",
        "currency",
        "delinquent",
        "default_source",
        subscription_status,
        "date_purged",
        "stripe_account",
    ]
    list_filter = [
        "delinquent",
        CustomerHasCardListFilter,
        CustomerSubscriptionStatusListFilter,
        AccountListFilter,
    ]
    search_fields = [
        "stripe_id",
    ] + user_search_fields()
    inlines = [
        SubscriptionInline,
        CardInline,
        BitcoinReceiverInline
    ]


class InvoiceItemInline(admin.TabularInline):
    model = InvoiceItem
    extra = 0
    max_num = 0


def customer_has_card(obj):
    return obj.customer.card_set.exclude(fingerprint="").exists()
customer_has_card.short_description = "Customer Has Card"  # noqa


def customer_user(obj):
    if not obj.customer.user:
        return ""
    User = get_user_model()
    username = getattr(obj.customer.user, User.USERNAME_FIELD)
    email = getattr(obj, "email", "")
    return "{0} <{1}>".format(
        username,
        email
    )
customer_user.short_description = "Customer"  # noqa


class InvoiceAdmin(ModelAdmin):
    raw_id_fields = ["customer"]
    list_display = [
        "stripe_id",
        "paid",
        "closed",
        customer_user,
        customer_has_card,
        "period_start",
        "period_end",
        "subtotal",
        "total"
    ]
    search_fields = [
        "stripe_id",
        "customer__stripe_id",
    ] + customer_search_fields()
    list_filter = [
        InvoiceCustomerHasCardListFilter,
        "paid",
        "closed",
        "attempted",
        "attempt_count",
        "created_at",
        "date",
        "period_end",
        "total"
    ]
    inlines = [
        InvoiceItemInline
    ]
    readonly_fields = [
        "stripe_account_stripe_id",
    ]


class PlanAdmin(ModelAdmin):
    raw_id_fields = ["stripe_account"]
    list_display = [
        "stripe_id",
        "name",
        "amount",
        "currency",
        "interval",
        "interval_count",
        "trial_period_days",
        "stripe_account",
    ]
    search_fields = [
        "stripe_id",
        "name",
        "=stripe_account__stripe_id",
    ] + customer_search_fields()
    list_filter = [
        "currency",
        AccountListFilter,
    ]


class CouponAdmin(ModelAdmin):
    list_display = [
        "stripe_id",
        "amount_off",
        "currency",
        "percent_off",
        "duration",
        "duration_in_months",
        "redeem_by",
        "valid"
    ]
    search_fields = [
        "stripe_id",
    ]
    list_filter = [
        "currency",
        "valid",
    ]


class TransferChargeFeeInline(admin.TabularInline):
    model = TransferChargeFee
    extra = 0
    max_num = 0


class TransferAdmin(ModelAdmin):
    Transfer
    raw_id_fields = ["event", "stripe_account"]
    list_display = [
        "stripe_id",
        "amount",
        "status",
        "date",
        "description",
        "stripe_account",
    ]
    search_fields = [
        "stripe_id",
        "event__stripe_id",
        "=stripe_account__stripe_id",
    ]
    inlines = [
        TransferChargeFeeInline
    ]
    list_filter = [
        AccountListFilter,
    ]


class AccountAdmin(ModelAdmin):
    raw_id_fields = ["user"]
    list_display = [
        "display_name",
        "type",
        "country",
        "payouts_enabled",
        "charges_enabled",
        "stripe_id",
        "created_at",
    ]
    search_fields = [
        "display_name",
        "stripe_id",
    ]


class BankAccountAdmin(ModelAdmin):
    raw_id_fields = ["account"]
    list_display = [
        "stripe_id",
        "account",
        "account_holder_type",
        "account_holder_name",
        "currency",
        "default_for_currency",
        "bank_name",
        "country",
        "last4"
    ]
    search_fields = [
        "stripe_id",
    ]


class UserAccountAdmin(ModelAdmin):
    raw_id_fields = ["user", "customer"]
    list_display = ["user", "customer"]
    search_fields = [
        "=customer__stripe_id",
        "=user__email",
    ]


admin.site.register(Account, AccountAdmin)
admin.site.register(BankAccount, BankAccountAdmin)
admin.site.register(Charge, ChargeAdmin)
admin.site.register(Coupon, CouponAdmin)
admin.site.register(Event, EventAdmin)
admin.site.register(EventProcessingException, EventProcessingExceptionAdmin)
admin.site.register(Invoice, InvoiceAdmin)
admin.site.register(Customer, CustomerAdmin)
admin.site.register(Plan, PlanAdmin)
admin.site.register(UserAccount, UserAccountAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.10 on 2016-12-20 03:16
from __future__ import unicode_literals

import django.utils.timezone
from django.db import migrations, models

import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0005_auto_20161006_1445'),
    ]

    operations = [
        migrations.CreateModel(
            name='Coupon',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('stripe_id', models.CharField(max_length=191, unique=True)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('amount_off', models.DecimalField(decimal_places=2, max_digits=9, null=True)),
                ('currency', models.CharField(default='usd', max_length=10)),
                ('duration', models.CharField(default='once', max_length=10)),
                ('duration_in_months', models.PositiveIntegerField(null=True)),
                ('livemode', models.BooleanField(default=False)),
                ('max_redemptions', models.PositiveIntegerField(null=True)),
                ('metadata', jsonfield.fields.JSONField(null=True)),
                ('percent_off', models.PositiveIntegerField(null=True)),
                ('redeem_by', models.DateTimeField(null=True)),
                ('times_redeemed', models.PositiveIntegerField(null=True)),
                ('valid', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import datetime

import stripe

from .. import models, utils
from .externalaccounts import sync_bank_account_from_stripe_data


def create(user, country, **kwargs):
    """
    Create an Account.

    Args:
        country: two letter country code for where the individual lives

    Returns:
        a pinax.stripe.models.Account object
    """
    kwargs["country"] = country
    stripe_account = stripe.Account.create(**kwargs)
    return sync_account_from_stripe_data(
        stripe_account, user=user
    )


def update(account, data):
    """
    Update the given account with extra data.

    Args:
        account: a pinax.stripe.models.Account object
        data: dict of account fields to update via API:

            first_name         -> legal_entity.first_name
            last_name          -> legal_entity.last_name
            dob                -> legal_entity.dob
            personal_id_number -> legal_entity.personal_id_number
            document           -> legal_entity.verification.document

    Returns:
        a pinax.stripe.models.Account object
    """
    stripe_account = stripe.Account.retrieve(id=account.stripe_id)

    if data.get("dob"):
        stripe_account.legal_entity.dob = data["dob"]

    if data.get("first_name"):
        stripe_account.legal_entity.first_name = data["first_name"]

    if data.get("last_name"):
        stripe_account.legal_entity.last_name = data["last_name"]

    if data.get("personal_id_number"):
        stripe_account.legal_entity.personal_id_number = data["personal_id_number"]

    if data.get("document"):
        response = stripe.FileUpload.create(
            purpose="identity_document",
            file=data["document"],
            stripe_account=stripe_account.id
        )
        stripe_account.legal_entity.verification.document = response["id"]

    stripe_account.save()
    return sync_account_from_stripe_data(stripe_account)


def sync_account(account):
    """
    Update the given local Account instance from remote data.

    Args:
        account: a pinax.stripe.models.Account object

    Returns:
        a pinax.stripe.models.Account object
    """
    stripe_account = stripe.Account.retrieve(id=account.stripe_id)
    return sync_account_from_stripe_data(stripe_account)


def sync_account_from_stripe_data(data, user=None):
    """
    Create or update using the account object from a Stripe API query.

    Args:
        data: the data representing an account object in the Stripe API

    Returns:
        a pinax.stripe.models.Account object
    """
    kwargs = {"stripe_id": data["id"]}
    if user:
        kwargs["user"] = user
    obj, created = models.Account.objects.get_or_create(
        **kwargs
    )
    common_attrs = (
        "business_name", "business_url", "charges_enabled", "country",
        "default_currency", "details_submitted", "display_name",
        "email", "type", "statement_descriptor", "support_email",
        "support_phone", "timezone", "payouts_enabled"
    )

    custom_attrs = (
        "debit_negative_balances", "metadata", "product_description",
        "payout_statement_descriptor"
    )

    if data["type"] == "custom":
        top_level_attrs = common_attrs + custom_attrs
    else:
        top_level_attrs = common_attrs

    for a in [x for x in top_level_attrs if x in data]:
        setattr(obj, a, data[a])

    # that's all we get for standard and express accounts!
    if data["type"] != "custom":
        obj.save()
        return obj

    # otherwise we continue on to gather a range of details available
    # to us on custom accounts

    # legal entity for individual accounts
    le = data["legal_entity"]
    address = le["address"]
    obj.legal_entity_address_city = address["city"]
    obj.legal_entity_address_country = address["country"]
    obj.legal_entity_address_line1 = address["line1"]
    obj.legal_entity_address_line2 = address["line2"]
    obj.legal_entity_address_postal_code = address["postal_code"]
    obj.legal_entity_address_state = address["state"]

    dob = le["dob"]
    if dob:
        obj.legal_entity_dob = datetime.date(
            dob["year"], dob["month"], dob["day"]
        )
    else:
        obj.legal_entity_dob = None

    obj.legal_entity_type = le["type"]
    obj.legal_entity_first_name = le["first_name"]
    obj.legal_entity_last_name = le["last_name"]
    obj.legal_entity_personal_id_number_provided = le["personal_id_number_provided"]

    # these attributes are not always present
    obj.legal_entity_gender = le.get(
        "gender", obj.legal_entity_gender
    )
    obj.legal_entity_maiden_name = le.get(
        "maiden_name", obj.legal_entity_maiden_name
    )
    obj.legal_entity_phone_number = le.get(
        "phone_number", obj.legal_entity_phone_number
    )
    obj.legal_entity_ssn_last_4_provided = le.get(
        "ssn_last_4_provided", obj.legal_entity_ssn_last_4_provided
    )

    verification = le["verification"]
    if verification:
        obj.legal_entity_verification_details = verification.get("details")
        obj.legal_entity_verification_details_code = verification.get("details_code")
        obj.legal_entity_verification_document = verification.get("document")
        obj.legal_entity_verification_status = verification.get("status")
    else:
        obj.legal_entity_verification_details = None
        obj.legal_entity_verification_details_code = None
        obj.legal_entity_verification_document = None
        obj.legal_entity_verification_status = None

    # tos state
    if data["tos_acceptance"]["date"]:
        obj.tos_acceptance_date = datetime.datetime.utcfromtimestamp(
            data["tos_acceptance"]["date"]
        )
    else:
        obj.tos_acceptance_date = None
    obj.tos_acceptance_ip = data["tos_acceptance"]["ip"]
    obj.tos_acceptance_user_agent = data["tos_acceptance"]["user_agent"]

    # decline charge on certain conditions
    obj.decline_charge_on_avs_failure = data["decline_charge_on"]["avs_failure"]
    obj.decline_charge_on_cvc_failure = data["decline_charge_on"]["cvc_failure"]

    # transfer schedule to external account
    ps = data["payout_schedule"]
    obj.payout_schedule_interval = ps["interval"]
    obj.payout_schedule_delay_days = ps.get("delay_days")
    obj.payout_schedule_weekly_anchor = ps.get("weekly_anchor")
    obj.payout_schedule_monthly_anchor = ps.get("monthly_anchor")

    # verification status, key to progressing account setup
    obj.verification_disabled_reason = data["verification"]["disabled_reason"]
    obj.verification_due_by = utils.convert_tstamp(data["verification"], "due_by")
    obj.verification_fields_needed = data["verification"]["fields_needed"]

    obj.save()

    # sync any external accounts (bank accounts only for now) included
    for external_account in data["external_accounts"]["data"]:
        if external_account["object"] == "bank_account":
            sync_bank_account_from_stripe_data(external_account)

    return obj


def delete(account):
    """
    Delete an account both remotely and locally.

    Note that this will fail if the account's balance is
    non-zero.
    """
    account.stripe_account.delete()
    account.delete()


def deauthorize(account):
    account.authorized = False
    account.save()



import stripe

from .. import models, utils


def sync_coupons():
    """
    Synchronizes all coupons from the Stripe API
    """
    coupons = stripe.Coupon.auto_paging_iter()
    for coupon in coupons:
        defaults = dict(
            amount_off=(
                utils.convert_amount_for_db(coupon["amount_off"], coupon["currency"])
                if coupon["amount_off"]
                else None
            ),
            currency=coupon["currency"] or "",
            duration=coupon["duration"],
            duration_in_months=coupon["duration_in_months"],
            max_redemptions=coupon["max_redemptions"],
            metadata=coupon["metadata"],
            percent_off=coupon["percent_off"],
            redeem_by=utils.convert_tstamp(coupon["redeem_by"]) if coupon["redeem_by"] else None,
            times_redeemed=coupon["times_redeemed"],
            valid=coupon["valid"],
        )
        obj, created = models.Coupon.objects.get_or_create(
            stripe_id=coupon["id"],
            defaults=defaults
        )
        utils.update_with_defaults(obj, defaults, created)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2017-08-25 18:41
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0008_auto_20170509_1736'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transfer',
            name='event',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='transfers', to='pinax_stripe.Event'),
        ),
    ]



from django.template import Context, Template
from django.test import TestCase, override_settings


class TemplateTagTests(TestCase):

    @override_settings(PINAX_STRIPE_PUBLIC_KEY="this-is-the-stripe-public-key")
    def test_stripe_public_key(self):
        self.maxDiff = None
        template = Template("""{% load stripe %}
            <script>
                Stripe.setPublishableKey({% stripe_public_key %});
            </script>
            """)

        self.assertEqual(
            template.render(Context()),
            """
            <script>
                Stripe.setPublishableKey('this-is-the-stripe-public-key');
            </script>
            """
        )

    def test_no_stripe_public_key(self):
        self.maxDiff = None
        template = Template("""{% load stripe %}
            <script>
                Stripe.setPublishableKey({% stripe_public_key %});
            </script>
            """)

        self.assertEqual(
            template.render(Context()),
            """
            <script>
                Stripe.setPublishableKey(*** PINAX_STRIPE_PUBLIC_KEY NOT SET ***);
            </script>
            """
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import decimal

from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

import stripe
from jsonfield.fields import JSONField

from .conf import settings
from .managers import ChargeManager, CustomerManager
from .utils import CURRENCY_SYMBOLS


class StripeObject(models.Model):

    stripe_id = models.CharField(max_length=191, unique=True)
    created_at = models.DateTimeField(default=timezone.now)

    class Meta:
        abstract = True


class AccountRelatedStripeObjectMixin(models.Model):

    stripe_account = models.ForeignKey(
        "pinax_stripe.Account",
        on_delete=models.CASCADE,
        null=True,
        default=None,
        blank=True,
    )

    @property
    def stripe_account_stripe_id(self):
        return getattr(self.stripe_account, "stripe_id", None)
    stripe_account_stripe_id.fget.short_description = "Stripe Account"

    class Meta:
        abstract = True


class AccountRelatedStripeObject(AccountRelatedStripeObjectMixin, StripeObject):
    """Uses a mixin to support Django 1.8 (name clash for stripe_id)"""

    class Meta:
        abstract = True


class UniquePerAccountStripeObject(AccountRelatedStripeObjectMixin):
    stripe_id = models.CharField(max_length=191)
    created_at = models.DateTimeField(default=timezone.now)

    class Meta:
        abstract = True
        unique_together = ("stripe_id", "stripe_account")


class StripeAccountFromCustomerMixin(object):
    @property
    def stripe_account(self):
        customer = getattr(self, "customer", None)
        return customer.stripe_account if customer else None

    @property
    def stripe_account_stripe_id(self):
        return self.stripe_account.stripe_id if self.stripe_account else None
    stripe_account_stripe_id.fget.short_description = "Stripe Account"


@python_2_unicode_compatible
class Plan(UniquePerAccountStripeObject):
    amount = models.DecimalField(decimal_places=2, max_digits=9)
    currency = models.CharField(max_length=15, blank=False)
    interval = models.CharField(max_length=15)
    interval_count = models.IntegerField()
    name = models.CharField(max_length=150)
    statement_descriptor = models.TextField(blank=True)
    trial_period_days = models.IntegerField(null=True, blank=True)
    metadata = JSONField(null=True, blank=True)

    def __str__(self):
        return "{} ({}{})".format(self.name, CURRENCY_SYMBOLS.get(self.currency, ""), self.amount)

    def __repr__(self):
        return "Plan(pk={!r}, name={!r}, amount={!r}, currency={!r}, interval={!r}, interval_count={!r}, trial_period_days={!r}, stripe_id={!r})".format(
            self.pk,
            self.name,
            self.amount,
            self.currency,
            self.interval,
            self.interval_count,
            self.trial_period_days,
            self.stripe_id,
        )

    @property
    def stripe_plan(self):
        return stripe.Plan.retrieve(
            self.stripe_id,
            stripe_account=self.stripe_account_stripe_id,
        )


@python_2_unicode_compatible
class Coupon(StripeObject):

    amount_off = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    currency = models.CharField(max_length=10, default="usd")
    duration = models.CharField(max_length=10, default="once")
    duration_in_months = models.PositiveIntegerField(null=True, blank=True)
    livemode = models.BooleanField(default=False)
    max_redemptions = models.PositiveIntegerField(null=True, blank=True)
    metadata = JSONField(null=True, blank=True)
    percent_off = models.PositiveIntegerField(null=True, blank=True)
    redeem_by = models.DateTimeField(null=True, blank=True)
    times_redeemed = models.PositiveIntegerField(null=True, blank=True)
    valid = models.BooleanField(default=False)

    def __str__(self):
        if self.amount_off is None:
            description = "{}% off".format(self.percent_off,)
        else:
            description = "{}{}".format(CURRENCY_SYMBOLS.get(self.currency, ""), self.amount_off)

        return "Coupon for {}, {}".format(description, self.duration)


@python_2_unicode_compatible
class EventProcessingException(models.Model):

    event = models.ForeignKey("Event", null=True, blank=True, on_delete=models.CASCADE)
    data = models.TextField()
    message = models.CharField(max_length=500)
    traceback = models.TextField()
    created_at = models.DateTimeField(default=timezone.now)

    def __str__(self):
        return "<{}, pk={}, Event={}>".format(self.message, self.pk, self.event)


@python_2_unicode_compatible
class Event(AccountRelatedStripeObject):

    kind = models.CharField(max_length=250)
    livemode = models.BooleanField(default=False)
    customer = models.ForeignKey("Customer", null=True, blank=True, on_delete=models.CASCADE)
    webhook_message = JSONField()
    validated_message = JSONField(null=True, blank=True)
    valid = models.NullBooleanField(null=True, blank=True)
    processed = models.BooleanField(default=False)
    request = models.CharField(max_length=100, blank=True)
    pending_webhooks = models.PositiveIntegerField(default=0)
    api_version = models.CharField(max_length=100, blank=True)

    @property
    def message(self):
        return self.validated_message

    def __str__(self):
        return "{} - {}".format(self.kind, self.stripe_id)

    def __repr__(self):
        return "Event(pk={!r}, kind={!r}, customer={!r}, valid={!r}, created_at={!s}, stripe_id={!r})".format(
            self.pk,
            self.kind,
            self.customer,
            self.valid,
            self.created_at.replace(microsecond=0).isoformat(),
            self.stripe_id,
        )


class Transfer(AccountRelatedStripeObject):

    amount = models.DecimalField(decimal_places=2, max_digits=9)
    amount_reversed = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    application_fee = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    created = models.DateTimeField(null=True, blank=True)
    currency = models.CharField(max_length=25, default="usd")
    date = models.DateTimeField()
    description = models.TextField(null=True, blank=True)
    destination = models.TextField(null=True, blank=True)
    destination_payment = models.TextField(null=True, blank=True)
    event = models.ForeignKey(
        Event, related_name="transfers",
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )
    failure_code = models.TextField(null=True, blank=True)
    failure_message = models.TextField(null=True, blank=True)
    livemode = models.BooleanField(default=False)
    metadata = JSONField(null=True, blank=True)
    method = models.TextField(null=True, blank=True)
    reversed = models.BooleanField(default=False)
    source_transaction = models.TextField(null=True, blank=True)
    source_type = models.TextField(null=True, blank=True)
    statement_descriptor = models.TextField(null=True, blank=True)
    status = models.CharField(max_length=25)
    transfer_group = models.TextField(null=True, blank=True)
    type = models.TextField(null=True, blank=True)

    @property
    def stripe_transfer(self):
        return stripe.Transfer.retrieve(
            self.stripe_id,
            stripe_account=self.stripe_account_stripe_id,
        )


class TransferChargeFee(models.Model):

    transfer = models.ForeignKey(Transfer, related_name="charge_fee_details", on_delete=models.CASCADE)
    amount = models.DecimalField(decimal_places=2, max_digits=9)
    currency = models.CharField(max_length=10, default="usd")
    application = models.TextField(null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    kind = models.CharField(max_length=150)
    created_at = models.DateTimeField(default=timezone.now)


class UserAccount(models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL,
                             related_name="user_accounts",
                             related_query_name="user_account",
                             on_delete=models.CASCADE)
    account = models.ForeignKey("pinax_stripe.Account",
                                related_name="user_accounts",
                                related_query_name="user_account",
                                on_delete=models.CASCADE)
    customer = models.ForeignKey("pinax_stripe.Customer",
                                 related_name="user_accounts",
                                 related_query_name="user_account",
                                 on_delete=models.CASCADE)

    class Meta:
        unique_together = ("user", "account")

    def clean(self):
        if not self.customer.stripe_account == self.account:
            raise ValidationError(_("customer.stripe_account must be account."))
        return super(UserAccount, self).clean()

    def save(self, *args, **kwargs):
        self.full_clean()
        return super(UserAccount, self).save(*args, **kwargs)

    def __repr__(self):
        return "UserAccount(pk={self.pk!r}, user={self.user!r}, account={self.account!r}, customer={self.customer!r})".format(self=self)


@python_2_unicode_compatible
class Customer(AccountRelatedStripeObject):

    user = models.OneToOneField(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE)
    users = models.ManyToManyField(settings.AUTH_USER_MODEL, through=UserAccount,
                                   related_name="customers",
                                   related_query_name="customers")
    account_balance = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    currency = models.CharField(max_length=10, default="usd", blank=True)
    delinquent = models.BooleanField(default=False)
    default_source = models.TextField(blank=True)
    date_purged = models.DateTimeField(null=True, blank=True, editable=False)

    objects = CustomerManager()

    @cached_property
    def stripe_customer(self):
        return stripe.Customer.retrieve(
            self.stripe_id,
            stripe_account=self.stripe_account_stripe_id,
        )

    def __str__(self):
        if self.user:
            return str(self.user)
        elif self.id:
            users = self.users.all()
            if users:
                return ", ".join(str(user) for user in users)
        if self.stripe_id:
            return "No User(s) ({})".format(self.stripe_id)
        return "No User(s)"

    def __repr__(self):
        if self.user:
            return "Customer(pk={!r}, user={!r}, stripe_id={!r})".format(
                self.pk,
                self.user,
                self.stripe_id,
            )
        elif self.id:
            return "Customer(pk={!r}, users={}, stripe_id={!r})".format(
                self.pk,
                ", ".join(repr(user) for user in self.users.all()),
                self.stripe_id,
            )
        return "Customer(pk={!r}, stripe_id={!r})".format(self.pk, self.stripe_id)


class Card(StripeObject):

    customer = models.ForeignKey(Customer, on_delete=models.CASCADE)
    name = models.TextField(blank=True)
    address_line_1 = models.TextField(blank=True)
    address_line_1_check = models.CharField(max_length=15)
    address_line_2 = models.TextField(blank=True)
    address_city = models.TextField(blank=True)
    address_state = models.TextField(blank=True)
    address_country = models.TextField(blank=True)
    address_zip = models.TextField(blank=True)
    address_zip_check = models.CharField(max_length=15)
    brand = models.TextField(blank=True)
    country = models.CharField(max_length=2, blank=True)
    cvc_check = models.CharField(max_length=15, blank=True)
    dynamic_last4 = models.CharField(max_length=4, blank=True)
    tokenization_method = models.CharField(max_length=15, blank=True)
    exp_month = models.IntegerField()
    exp_year = models.IntegerField()
    funding = models.CharField(max_length=15)
    last4 = models.CharField(max_length=4, blank=True)
    fingerprint = models.TextField()

    def __repr__(self):
        return "Card(pk={!r}, customer={!r})".format(
            self.pk,
            getattr(self, "customer", None),
        )


class BitcoinReceiver(StripeObject):

    customer = models.ForeignKey(Customer, on_delete=models.CASCADE)
    active = models.BooleanField(default=False)
    amount = models.DecimalField(decimal_places=2, max_digits=9)
    amount_received = models.DecimalField(decimal_places=2, max_digits=9, default=decimal.Decimal("0"))
    bitcoin_amount = models.PositiveIntegerField()  # Satoshi (10^8 Satoshi in one bitcoin)
    bitcoin_amount_received = models.PositiveIntegerField(default=0)
    bitcoin_uri = models.TextField(blank=True)
    currency = models.CharField(max_length=10, default="usd")
    description = models.TextField(blank=True)
    email = models.TextField(blank=True)
    filled = models.BooleanField(default=False)
    inbound_address = models.TextField(blank=True)
    payment = models.TextField(blank=True)
    refund_address = models.TextField(blank=True)
    uncaptured_funds = models.BooleanField(default=False)
    used_for_payment = models.BooleanField(default=False)


class Subscription(StripeAccountFromCustomerMixin, StripeObject):

    STATUS_CURRENT = ["trialing", "active"]

    customer = models.ForeignKey(Customer, on_delete=models.CASCADE)
    application_fee_percent = models.DecimalField(decimal_places=2, max_digits=3, default=None, null=True, blank=True)
    cancel_at_period_end = models.BooleanField(default=False)
    canceled_at = models.DateTimeField(null=True, blank=True)
    current_period_end = models.DateTimeField(null=True, blank=True)
    current_period_start = models.DateTimeField(null=True, blank=True)
    ended_at = models.DateTimeField(null=True, blank=True)
    plan = models.ForeignKey(Plan, on_delete=models.CASCADE)
    quantity = models.IntegerField()
    start = models.DateTimeField()
    status = models.CharField(max_length=25)  # trialing, active, past_due, canceled, or unpaid
    trial_end = models.DateTimeField(null=True, blank=True)
    trial_start = models.DateTimeField(null=True, blank=True)

    @property
    def stripe_subscription(self):
        return stripe.Subscription.retrieve(self.stripe_id, stripe_account=self.stripe_account_stripe_id)

    @property
    def total_amount(self):
        return self.plan.amount * self.quantity

    def plan_display(self):
        return self.plan.name

    def status_display(self):
        return self.status.replace("_", " ").title()

    def delete(self, using=None):
        """
        Set values to None while deleting the object so that any lingering
        references will not show previous values (such as when an Event
        signal is triggered after a subscription has been deleted)
        """
        super(Subscription, self).delete(using=using)
        self.status = None
        self.quantity = 0
        self.amount = 0

    def __repr__(self):
        return "Subscription(pk={!r}, customer={!r}, plan={!r}, status={!r}, stripe_id={!r})".format(
            self.pk,
            getattr(self, "customer", None),
            getattr(self, "plan", None),
            self.status,
            self.stripe_id,
        )


class Invoice(StripeAccountFromCustomerMixin, StripeObject):

    customer = models.ForeignKey(Customer, related_name="invoices", on_delete=models.CASCADE)
    amount_due = models.DecimalField(decimal_places=2, max_digits=9)
    attempted = models.NullBooleanField()
    attempt_count = models.PositiveIntegerField(null=True, blank=True)
    charge = models.ForeignKey("Charge", null=True, blank=True, related_name="invoices", on_delete=models.CASCADE)
    subscription = models.ForeignKey(Subscription, null=True, blank=True, on_delete=models.CASCADE)
    statement_descriptor = models.TextField(blank=True)
    currency = models.CharField(max_length=10, default="usd")
    closed = models.BooleanField(default=False)
    description = models.TextField(blank=True)
    paid = models.BooleanField(default=False)
    receipt_number = models.TextField(blank=True)
    period_end = models.DateTimeField()
    period_start = models.DateTimeField()
    subtotal = models.DecimalField(decimal_places=2, max_digits=9)
    tax = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    tax_percent = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    total = models.DecimalField(decimal_places=2, max_digits=9)
    date = models.DateTimeField()
    webhooks_delivered_at = models.DateTimeField(null=True, blank=True)

    @property
    def status(self):
        return "Paid" if self.paid else "Open"

    @property
    def stripe_invoice(self):
        return stripe.Invoice.retrieve(
            self.stripe_id,
            stripe_account=self.stripe_account_stripe_id,
        )


class InvoiceItem(models.Model):

    stripe_id = models.CharField(max_length=255)
    created_at = models.DateTimeField(default=timezone.now)
    invoice = models.ForeignKey(Invoice, related_name="items", on_delete=models.CASCADE)
    amount = models.DecimalField(decimal_places=2, max_digits=9)
    currency = models.CharField(max_length=10, default="usd")
    kind = models.CharField(max_length=25, blank=True)
    subscription = models.ForeignKey(Subscription, null=True, blank=True, on_delete=models.CASCADE)
    period_start = models.DateTimeField()
    period_end = models.DateTimeField()
    proration = models.BooleanField(default=False)
    line_type = models.CharField(max_length=50)
    description = models.CharField(max_length=200, blank=True)
    plan = models.ForeignKey(Plan, null=True, blank=True, on_delete=models.CASCADE)
    quantity = models.IntegerField(null=True, blank=True)

    def plan_display(self):
        return self.plan.name if self.plan else ""


class Charge(StripeAccountFromCustomerMixin, StripeObject):

    customer = models.ForeignKey(Customer, null=True, blank=True, related_name="charges", on_delete=models.CASCADE)
    invoice = models.ForeignKey(Invoice, null=True, blank=True, related_name="charges", on_delete=models.CASCADE)
    source = models.CharField(max_length=100, blank=True)
    currency = models.CharField(max_length=10, default="usd")
    amount = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    amount_refunded = models.DecimalField(decimal_places=2, max_digits=9, null=True, blank=True)
    description = models.TextField(blank=True)
    paid = models.NullBooleanField(null=True, blank=True)
    disputed = models.NullBooleanField(null=True, blank=True)
    refunded = models.NullBooleanField(null=True, blank=True)
    captured = models.NullBooleanField(null=True, blank=True)
    receipt_sent = models.BooleanField(default=False)
    charge_created = models.DateTimeField(null=True, blank=True)

    # These fields are extracted from the BalanceTransaction for the
    # charge and help us to know when funds from a charge are added to
    # our Stripe account's balance.
    available = models.BooleanField(default=False)
    available_on = models.DateTimeField(null=True, blank=True)
    fee = models.DecimalField(
        decimal_places=2, max_digits=9, null=True, blank=True
    )
    fee_currency = models.CharField(max_length=10, null=True, blank=True)

    transfer_group = models.TextField(null=True, blank=True)
    outcome = JSONField(null=True, blank=True)

    objects = ChargeManager()

    def __str__(self):
        info = []
        if not self.paid:
            info += ["unpaid"]
        if not self.captured:
            info += ["uncaptured"]
        if self.refunded:
            info += ["refunded"]
        currency = CURRENCY_SYMBOLS.get(self.currency, "")
        return "{}{}{}".format(
            currency,
            self.total_amount,
            " ({})".format(", ".join(info)) if info else "",
        )

    def __repr__(self):
        return "Charge(pk={!r}, customer={!r}, source={!r}, amount={!r}, captured={!r}, paid={!r}, stripe_id={!r})".format(
            self.pk,
            self.customer,
            self.source,
            self.amount,
            self.captured,
            self.paid,
            self.stripe_id,
        )

    @property
    def total_amount(self):
        amount = self.amount if self.amount else 0
        amount_refunded = self.amount_refunded if self.amount_refunded else 0
        return amount - amount_refunded
    total_amount.fget.short_description = "Σ amount"

    @property
    def stripe_charge(self):
        return stripe.Charge.retrieve(
            self.stripe_id,
            stripe_account=self.stripe_account_stripe_id,
            expand=["balance_transaction"]
        )

    @property
    def card(self):
        return Card.objects.filter(stripe_id=self.source).first()


@python_2_unicode_compatible
class Account(StripeObject):

    INTERVAL_CHOICES = (
        ("Manual", "manual"),
        ("Daily", "daily"),
        ("Weekly", "weekly"),
        ("Monthly", "monthly"),
    )
    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, on_delete=models.CASCADE, related_name="stripe_accounts")

    business_name = models.TextField(null=True, blank=True)
    business_url = models.TextField(null=True, blank=True)

    charges_enabled = models.BooleanField(default=False)
    country = models.CharField(max_length=2)
    debit_negative_balances = models.BooleanField(default=False)
    decline_charge_on_avs_failure = models.BooleanField(default=False)
    decline_charge_on_cvc_failure = models.BooleanField(default=False)
    default_currency = models.CharField(max_length=3)
    details_submitted = models.BooleanField(default=False)
    display_name = models.TextField(blank=True, null=True)
    email = models.TextField(null=True, blank=True)

    legal_entity_address_city = models.TextField(null=True, blank=True)
    legal_entity_address_country = models.TextField(null=True, blank=True)
    legal_entity_address_line1 = models.TextField(null=True, blank=True)
    legal_entity_address_line2 = models.TextField(null=True, blank=True)
    legal_entity_address_postal_code = models.TextField(null=True, blank=True)
    legal_entity_address_state = models.TextField(null=True, blank=True)
    legal_entity_dob = models.DateField(null=True, blank=True)
    legal_entity_first_name = models.TextField(null=True, blank=True)
    legal_entity_gender = models.TextField(null=True, blank=True)
    legal_entity_last_name = models.TextField(null=True, blank=True)
    legal_entity_maiden_name = models.TextField(null=True, blank=True)
    legal_entity_personal_id_number_provided = models.BooleanField(default=False)
    legal_entity_phone_number = models.TextField(null=True, blank=True)
    legal_entity_ssn_last_4_provided = models.BooleanField(default=False)
    legal_entity_type = models.TextField(null=True, blank=True)
    legal_entity_verification_details = models.TextField(null=True, blank=True)
    legal_entity_verification_details_code = models.TextField(null=True, blank=True)
    legal_entity_verification_document = models.TextField(null=True, blank=True)
    legal_entity_verification_status = models.TextField(null=True, blank=True)

    # The type of the Stripe account. Can be "standard", "express", or "custom".
    type = models.TextField(null=True, blank=True)

    metadata = JSONField(null=True, blank=True)

    stripe_publishable_key = models.CharField(null=True, blank=True, max_length=100)

    product_description = models.TextField(null=True, blank=True)
    statement_descriptor = models.TextField(null=True, blank=True)
    support_email = models.TextField(null=True, blank=True)
    support_phone = models.TextField(null=True, blank=True)

    timezone = models.TextField(null=True, blank=True)

    tos_acceptance_date = models.DateField(null=True, blank=True)
    tos_acceptance_ip = models.TextField(null=True, blank=True)
    tos_acceptance_user_agent = models.TextField(null=True, blank=True)

    payout_schedule_delay_days = models.PositiveSmallIntegerField(null=True, blank=True)
    payout_schedule_interval = models.CharField(max_length=7, choices=INTERVAL_CHOICES, null=True, blank=True)
    payout_schedule_monthly_anchor = models.PositiveSmallIntegerField(null=True, blank=True)
    payout_schedule_weekly_anchor = models.TextField(null=True, blank=True)
    payout_statement_descriptor = models.TextField(null=True, blank=True)
    payouts_enabled = models.BooleanField(default=False)

    verification_disabled_reason = models.TextField(null=True, blank=True)
    verification_due_by = models.DateTimeField(null=True, blank=True)
    verification_timestamp = models.DateTimeField(null=True, blank=True)
    verification_fields_needed = JSONField(null=True, blank=True)
    authorized = models.BooleanField(default=True)

    @property
    def stripe_account(self):
        return stripe.Account.retrieve(self.stripe_id)

    def __str__(self):
        return "{} - {}".format(self.display_name or "", self.stripe_id)

    def __repr__(self):
        return "Account(pk={!r}, display_name={!r}, type={!r}, authorized={!r}, stripe_id={!r})".format(
            self.pk,
            self.display_name or "",
            self.type,
            self.authorized,
            self.stripe_id,
        )


class BankAccount(StripeObject):

    account = models.ForeignKey(Account, related_name="bank_accounts", on_delete=models.CASCADE)
    account_holder_name = models.TextField()
    account_holder_type = models.TextField()
    bank_name = models.TextField(null=True, blank=True)
    country = models.TextField()
    currency = models.TextField()
    default_for_currency = models.BooleanField(default=False)
    fingerprint = models.TextField()
    last4 = models.CharField(max_length=4)
    metadata = JSONField(null=True, blank=True)
    routing_number = models.TextField()
    status = models.TextField()

    @property
    def stripe_bankaccount(self):
        return self.account.stripe_account.external_accounts.retrieve(
            self.stripe_id
        )



# Generated by Django 2.0.4 on 2018-04-14 00:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0013_charge_outcome'),
    ]

    operations = [
        migrations.AlterField(
            model_name='account',
            name='display_name',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.6 on 2017-11-22 15:40
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('pinax_stripe', '0014_blank_with_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='charge',
            name='outcome',
            field=jsonfield.fields.JSONField(blank=True, null=True),
        ),
    ]



from .. import models


def create_bank_account(account, account_number, country, currency, **kwargs):
    """
    Create a Bank Account.

    Args:
        account: the stripe.Account object we're attaching
                 the bank account to
        account_number: the Bank Account number
        country: two letter country code
        currency: three letter currency code

        There are additional properties that can be set, please see:
        https://stripe.com/docs/api#account_create_bank_account

    Returns:
        a pinax.stripe.models.BankAccount object
    """
    external_account = account.external_accounts.create(
        external_account=dict(
            object="bank_account",
            account_number=account_number,
            country=country,
            currency=currency,
            **kwargs
        )
    )
    return sync_bank_account_from_stripe_data(
        external_account
    )


def sync_bank_account_from_stripe_data(data):
    """
    Create or update using the account object from a Stripe API query.

    Args:
        data: the data representing an account object in the Stripe API

    Returns:
        a pinax.stripe.models.Account object
    """
    account = models.Account.objects.get(
        stripe_id=data["account"]
    )
    kwargs = {
        "stripe_id": data["id"],
        "account": account
    }
    obj, created = models.BankAccount.objects.get_or_create(
        **kwargs
    )
    top_level_attrs = (
        "account_holder_name", "account_holder_type",
        "bank_name", "country", "currency", "default_for_currency",
        "fingerprint", "last4", "metadata", "routing_number",
        "status"
    )
    for a in top_level_attrs:
        setattr(obj, a, data.get(a))
    obj.save()
    return obj
